/*
 Program WinCaml: Graphical User Interface
 for interactive use of Caml-Light and Ocaml.
 Copyright (C) 2005-2024 Jean Mouric 35700 Rennes France
 email: jean.mouric@orange.fr
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

// File lex.tt.cpp

#include "platform.h"
#include "OCamlIndenter.h"

static size_t tokenbegin;
static size_t tokenend;

static int t;
static char* p;
static char* str;
static char buf[256];

extern size_t currentPos;
extern size_t token_length;
size_t ttleng;
extern int leading;
extern int trailing;
extern int newline;

static int d;
static size_t sz;

static const int keywordsMax = 56;
static const char* keywords[] = {"and", "as", "asr", "assert", "begin", "class", "constraint", "do", "done", "downto", "else", "end", "exception", "external", "false", "for", "fun", "function", "functor", "if", "in", "include", "inherit", "initializer", "land", "lazy", "let", "lnot", "lor", "lsl", "lsr", "lxor", "match", "method", "mod", "module", "mutable", "new", "object", "of", "open", "or", "parser", "private", "rec", "sig", "struct", "then", "to", "true", "try", "type", "val", "virtual", "when", "while", "with"};
static const int keyvalues[] = {AND, AS, INFIX_SYMBOL, ASSERT1, BEGIN1, CLASS, CONSTRAINT, DO, DONE, DOWNTO, ELSE, END, EXCEPTION, EXTERNAL, BOOLVALUE, FOR, FUN, FUNCTION, FUNCTOR, IF, IN1, INCLUDE, INHERIT, INITIALIZER, INFIX_SYMBOL, LAZY, LET, INFIX_SYMBOL, INFIX_SYMBOL, INFIX_SYMBOL, INFIX_SYMBOL, INFIX_SYMBOL, MATCH, METHOD1, INFIX_SYMBOL, MODULE, MUTABLE, NEW, OBJECT, OF, OPEN, INFIX_SYMBOL, PARSER, PRIVATE, REC, SIG, STRUCT, THEN, TO, BOOLVALUE, TRY1, TYPE, VAL, VIRTUAL, WHEN,
    WHILE, WITH};

static const int opCharsMax = 18;
static const char opChars[] = {'!', '$', '%', '&', '*', '+', '-', '.', '/', ':', ';', '<', '=', '>', '?', '@', '^', '|', '~'};

static const int opFirstCharsMax = 7;
static const char opFirstChars[] = {'$', '%', '&', '+', '/', '=', '@', '^'};

//------------------------------------------------------------------------------------------------------------------------------------

bool is_digit(char c);
static inline int ret(int tok, int lead, int trail);
static inline void whitespace();
static inline int keyvalue(char* buf);
static inline bool isOpChar(char c);
static inline bool isUpperCase(int c);
static inline bool isLowerCase(int c);
static inline bool isAlpha(int c);
static inline bool isAlNum(int c);

static inline int ret(int tok, int lead, int trail)
{
    currentPos = tokenend = (size_t)(p - str);
    if (tok == 0 || currentPos > sz) {
        currentPos = sz,
        leading = 0;
        trailing = 0;
        free(str);
        return 0;
    }
    token_length = ttleng = currentPos - tokenbegin;
    leading = lead;
    trailing = trail;
    return tok;
}

static inline void whitespace()
{
    while(*p == ' ' || *p == '\t' || *p == '\r' || *p == '\n') {
        if (*p == ' ' || *p == '\t') {
            p++;
        }
        else if (*p == '\n') {
            p++; newline++;
        }
        else {
            newline++;
            p++;
            if (*p == '\n') {
                p++;
            }
        }
    }
}

static inline int keyvalue(char* s)
{
    int l1 = 0;
    int l2 = keywordsMax;
    int m;
    if (strcmp(s, keywords[0]) < 0 || strcmp(keywords[keywordsMax], s) < 0) {
        return 0;
    }
    while (((m = (l1 + l2) / 2) != l1)) {
        if (strcmp(keywords[m], s) < 0) {
            l1 = m;
        }
        else if (strcmp(keywords[m], s) > 0){
            l2 = m;
        }
        else return keyvalues[m];
    }
    if (strcmp(keywords[l1], s) == 0) return keyvalues[l1];
    else if (strcmp(keywords[l2], s) == 0) return keyvalues[l2];
    else return 0;
}

static inline bool isOpChar(char c)
{
    int l1 = 0;
    int l2 = opCharsMax;
    int m;
    if (c < opChars[0] || c > opChars[opCharsMax]) {
        return false;
    }
    while (((m = (l1 + l2) / 2) != l1)) {
        if (opChars[m] < c) {
            l1 = m;
        }
        else if (opChars[m] > c){
            l2 = m;
        }
        else return true;
    }
    return opChars[l1] == c || opChars[l2] == c;
}

static inline bool isUpperCase(int c)
{
    return (64 < c && c < 91) || (191 < c && c < 215) || (215 < c && c < 223);
}

static inline bool isLowerCase(int c)
{
    return (96 < c && c < 123) || (222 < c && c < 247) || (248 < c && c < 256) || c == '_';
}

static inline bool isAlpha(int c)
{
    return isUpperCase(c) || isLowerCase(c);
}

static inline bool isAlNum(int c)
{
    return isAlpha(c) || is_digit(c);
}

//------------------------------------------------------------------------------------------------------------------------------------

static bool init = true;
static int (*tab[256])() = {NULL};
static  void inittab();
static  int lpar();
static  int rpar();
static  int comment();
static  int lbrace();
static  int rbrace();
static  int lbracket();
static  int rbracket();
static  int bigger();
static  int verticalbar();
static  int doublequote();
static  int backslash();
static  int semicolon();
static  int colon();
static  int minussign();
static  int underscore();
static  int sharp();
static  int comma();
static  int dot();
static  int charliteral();
static  int prefix();
static  int smaller();
static  int hexadecimal();
static  int num();
static  int opfirstchars();
static  int alpha();

static void inittab()
{
    tab[(int)'('] = lpar;
    tab[(int)')'] = rpar;
    tab[(int)'*'] = comment;
    tab[(int)'{'] = lbrace;
    tab[(int)'}'] = rbrace;
    tab[(int)'['] = lbracket;
    tab[(int)']'] = rbracket;
    tab[(int)'>'] = bigger;
    tab[(int)'|'] = verticalbar;
    tab[(int)'"'] = doublequote;
    tab[(int)'\\'] = backslash;
    tab[(int)';'] = semicolon;
    tab[(int)':'] = colon;
    tab[(int)'-'] = minussign;
    tab[(int)'_'] = underscore;
    tab[(int)'#'] = sharp;
    tab[(int)','] = comma;
    tab[(int)'.'] = dot;
    tab[(int)'\''] = tab[(int)'`'] = charliteral;
    tab[(int)'!'] = tab[(int)'?'] = tab[(int)'~'] = prefix;
    tab[(int)'<'] = smaller;
    tab[(int)'0'] = hexadecimal;
    int i;
    for (i = (int)'1'; i <= (int)'9'; i++) {
        tab[i] = num;
    }
    for (i = 0; i <=opFirstCharsMax; i++) {
        tab[(int)opFirstChars[i]] = opfirstchars;
    }
    for (i = 65; i < 256; i++) {
        if (isAlpha(i)) {
            tab[i] = alpha;
        }
    }
}

static int lpar()
{
    if (*p == '*') {
        d++;
        t = *(p + 1);
        p += 2;
        while (d != 0) {
            if (t == '\0') {
                p--;
                return ret(OTHER, 1, 0);
            }
            if (t == '(' && *p == '*') {
                d++;
                t = *(p + 1);
                p += 2;
            }
            else if (t == '*' && *p == ')') {
                d--;
                t = *(p + 1);
                p ++;
            }
            else t = *p++;
        }
        return ret(COMMENT, 1, 1);
    }
    char* r = p;
    while(*p == ' ' || *p == '\t' || *p == '\r' || *p == '\n') {
        p++;
    }
    if (*p == ')') {
        p++;
        return ret(UNIT, 1, 1);
    }
    p = r;
    return ret(LPAR, 1, 0);
}

static int rpar()
{
    return ret(RPAR, 0, 1);
}

static int comment()
{
    if (*p == ')') {
        p++;
        return ret(OTHER, 1, 1);
    }
    return opfirstchars();
}

static int lbrace(void) // quoted strings
{
    if (*p == '<') {
        p++;
        return ret(LOBJECT, 1, 1);
    }
    char* r = p;
    while (('a' <= *r && *r <= 'z') || *r == '_') { r++; } // quoted string start sequence ?
    if (*r == '|') { // if quoted string start sequence
        while (*r != '\0') { // as long as a character is available
            r++; // go forward (every character but '\0' is accepted)
            if (*r == '|') { // does this vertical bar begin the end of the quoted string ?...
                r++; // comparing to the quoted string start sequence ...
                char* r1 = p; // ...
                char* r2 = r; // ...
                while (*r1 == *r2) { r1++; r2++; } // ...
                if (*r2 == '}' && *r1 == '|') { p = ++r2; return ret(STRING, 1, 1); } // if the comparison was successful then return
                r = --r2; // else reposition r and restart as long as a character is available
            }
        }
        return ret(OTHER, 1, 0);
    }
    return ret(LBRACE, 1, 0);
}

static int rbrace()
{
    return ret(RBRACE, 0, 1);
}

static int lbracket()
{
    if (*p == '|') {
        p++;
        return ret(LLIST, 1, 1);
    }
    if (*p == '<') {
        p++;
        return ret(LSTREAM, 1, 1);
    }
    if (*p == '>') {
        p++;
        return ret(LBRACKETSUP, 1, 1);
    }
    char* r = p;
    while(*p == ' ' || *p == '\t' || *p == '\r' || *p == '\n') {
        p++;
    }
    if (*p == ']') {
        p++;
        return ret(EMPTYLIST, 1, 1);
    }
    p = r;
    return ret(LBRACKET, 1, 0);
}

static int rbracket()
{
    return ret(RBRACKET, 0, 1);
}

static int bigger()
{
    if (*p == '}') {
        p++;
        return ret(ROBJECT, 1, 1);
    }
    if (*p == ']') {
        p++;
        return ret(RSTREAM, 1, 1);
    }
    return opfirstchars();
}

static int verticalbar()
{
    if (*p == '|') {
        p++;
        return ret(INFIX_SYMBOL, 1, 1);
    }
    else if (*p == ']') {
        p++;
        return ret(RLIST,1, 1);
    }
    else {
        char c = *p;
        if (!isOpChar(c))
        {
            return ret(INFIX_SYMBOL, 1, 1);
        }
        return opfirstchars();
    }
}

int doublequote()
{
    char c = *p++;
    bool escape = true;
    while (c != '"' && c != '\0') {
        if (c == '\\' && *p == '"' && escape) {
            escape = false;
            c = *(p + 1);
            p += 2;
        }
        else if (c == '\\') {
            escape = !escape;
            c = *p++;
        }
        else {
            escape = true;
            c = *p++;
        }
    }
    if (c == '\0') {
        p--;
        return ret(OTHER, 1, 0);
    }
    return ret(STRING, 1, 1);
}

static int backslash()
{
    if (*p == '"') {
        p++;
    }
    return ret(OTHER, 1, 1);
}

static int semicolon()
{
    if (*p == ';') {
        p++;
        return ret(ENDMARK, 0, 0);
    }
    return ret(SEMICOLON, 0, 1);
}

static int colon()
{
    if (*p == ':') {
        p++;
        return ret(COLONCOLON, 1, 1);
    }
    if (*p == '=') {
        p++;
        return ret(INFIX_SYMBOL, 1, 1);
    }
    if (*p == '>') {
        p++;
        return ret(COLONSUP, 1, 1);
    }
    return ret(COLON, 0, 1);
}

static int minussign()
{
    if (*p == '>') {
        char c = *(p + 1);
        if (!isOpChar(c))
        {
            p++;
            return ret(RIGHTARROW, 1, 1);
        }
    }
    return opfirstchars();
}

static int underscore()
{
    char* q = p;
    int i = alpha();
    if (p > q) {
        return ret(i, 1, 1);
    }
    return ret(OTHER, 1, 1);
}

static int sharp()
{
    return ret(SHARP, 0, 0);
}

static int comma()
{
    return ret(COMMA, 0, 1);
}

static int dot()
{
    if (*p == '.') {
        p++;
        if (*p == '.') {
            p++;
            return ret(DOTDOTDOT, 1, 1);
        }
        return ret(DOTDOT, 1, 1);
    }
    return ret(DOT, 0, 0);
}

static int charliteral()
{
    if (*p == '\\') {
        p++;
        if (is_digit(*p) && is_digit(*(p + 1)) && is_digit(*(p + 2)) && *(p + 3) == t) {
            p += 4;
        }
        else if (*p == 'x' && isxdigit(*(p + 1)) && isxdigit(*(p + 2)) && *(p + 3) == t) {
            p += 4;
        }
        else if ((*p == 'n' || *p == 't' || *p == 'b' || *p == 'r' || *p == '\\' || *p == '\'' || *p == '`') && *(p + 1) == t) {
            p += 2;
        }
        return ret(CHAR, 1, 1);
    }
    if (*p == 'x' && *(p + 1) == 'x' && *(p + 2) == t ) {
        p += 3;
        return ret(CHAR, 1, 1);
    }
    if (*p == '\0') {
        return ret(OTHER,0 , 0);
    }
    if (*(p + 1) == t) {
        p += 2;
        return ret(CHAR, 1, 1);
    }
    return t == '`' ? ret(BACKQUOTE, 1, 0) : ret(QUOTE, 1, 1);
}

static int prefix()
{
    char* r = p;
    char c = *p;
    while (isOpChar(c))
    {
        c = *++p;
    }
    if (t == '!' && *r == '=' && p - r == 1) {
        return ret(INFIX_SYMBOL, 1, 1);
    }
    return ret(PREFIX_SYMBOL, 1, 0);
}

static int smaller()
{
    if (*p == '-') {
        char c = *(p + 1);
        if (!isOpChar(c))
        {
            p++;
            return ret(LEFTARROW, 1, 1);
        }
    }
    return opfirstchars();
}

static int hexadecimal()
{
    if (*p == 'x' || *p == 'X' || *p == 'o' || *p == 'O' || *p == 'b' || *p == 'B') {
        p++;
        while (isxdigit(*p) || *p == '_') {
            p++;
        }
        return ret(INTEGER, 1, 1);
    }
    return num();
}

static int num()
{
    bool isFloat = false;
    while (is_digit(*p)) {
        p++;
    }
    if (*p == '.') {
        isFloat = true;
        p++;
        while (is_digit(*p)) {
            p++;
        }
    }
    if (*p == 'e' || *p == 'E') {
        isFloat = true;
        p++;
        if (*p == '-' || *p == '+') {
            p++;
        }
        while (is_digit(*p)) {
            p++;
        }
    }
    return (isFloat ? ret(FLOAT, 1, 1) : ret(INTEGER, 1, 1));
}

static int opfirstchars()
{
    char c = *p;
    if (c) {
        char c1 = *(p + 1);
        while (c1 != ')' && isOpChar(c))
        {
            if (c == ';' && c1 == ';') {
                break;
            }
            c = *++p;
            c1 = *(p + 1);
        }
        if  (c1 == ')' && c != '*' && isOpChar(c)) p++;
    }
    return ret(INFIX_SYMBOL, 1, 1);
}

static int alpha()
{
    int i = 0;
    int c = (unsigned char)t;
    buf[i++] = t;
    while (isAlNum((unsigned char)*p) || *p == '_' || *p == '\'') {
        t = *p++;
        if (i < 255) buf[i++] = t;
    }
    buf[i] = 0;
    int kv = keyvalue(buf);
    if (kv != 0) {
        if (*p == '!' && (kv == INHERIT || kv == METHOD1 || kv == VAL)) {
            p++;
        }
        return ret(kv, 1, 1);
    }
    return ret(isLowerCase(c) ? LOWERCASE_IDENT : CAPITALIZED_IDENT, 1 ,1);
}

//------------------------------------------------------------------------------------------------------------------------------------

typedef struct yy_buffer_state *BUFFER_STATE;
BUFFER_STATE tt_scan_string (const char* s);
int ttlex();

BUFFER_STATE tt_scan_string (const char* s)
{
    if (init) {
        inittab();
        init = false;
    }
    sz = strlen(s);
    str = (char*)malloc(++sz);
    memcpy(str, s, sz--);
    tokenbegin = 0;
    tokenend = 0;
    token_length = ttleng = 0;
    newline = 0;
    d = 0;
    p = (char*)str;
    return NULL;
}

int ttlex()
{
    whitespace();
    tokenbegin = p - str;
    t = (int)(unsigned char)*p++;
    int (*f)() = tab[t];
    return f ? f() : ret(OTHER, 1, 1);
}
