/*
 Program WinCaml: Graphical User Interface
 for interactive use of Caml-Light and Ocaml.
 Copyright (C) 2005-2024 Jean Mouric 35700 Rennes France
 email: jean.mouric@orange.fr
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

// File Qt/Paths.cpp

#include "Qt.h"

#ifndef Q_OS_WIN
#include <stdio.h>
#include <sys/stat.h>
#include <iostream>
#include <fstream>
#include <filesystem> // nécessite -std=c++17

string fromUnicode(const wstring& wstr, int code); // définie dans Qt.cpp (uff16 vers code)
wstring toUnicode(const string& str, int code); // définie dans Qt.cpp (code vers utf-16) 
string toBytes(const wstring&); // utf-16 vers utf-8 (code = utf-8 ci-dessus)
wstring getCamlPath(const wstring&); // récupère le chemin de Caml Light ou OCaml inscrit sur disque 
bool resetCamlPath(const wstring&);
bool createDirectory(const wstring& dirPath);

extern bool camlnew;
extern CMDIFrame* mainFrame;

QString applicationDirPath()
{
#ifdef Q_OS_MAC
    QDir qd(qApp->applicationDirPath());
    qd.cdUp(); qd.cdUp(); qd.cdUp();
    return qd.path();
#else
    return qApp->applicationDirPath();
#endif
}
bool getConfigFolderPath(wstring& path) // path se termine par un "/"
{
    path = (applicationDirPath() + "/." + QDir(QDir::homePath()).dirName() + "/").toStdWString();
    return true;
}
wstring configFolderPath(void) // le chemin produit se termine par un "/"
{
	wstring path = L"";
	getConfigFolderPath(path);
	return path;
}

wstring getCamlPath(const wstring& camlbase) // camlbase est L"camlBase.txt" ou L"ocamlBase.txt"
{
    wstring cfp;
    string camlPath;
    if (getConfigFolderPath(cfp))
    {
        cfp += camlbase;
        fstream fs(toBytes(cfp).c_str(), fstream::in);
        if (!fs.fail())
        {
            char c;
            while (fs.get(c) && c != L'\r' && c != L'\n')
                camlPath += c;
            fs.close();
        }
    }
    return toUnicode(camlPath, ENCODING_UTF8);
}

void writeCamlPath(wstring camldirPath, wstring camlBase) // camlbase est L"camlBase.txt" ou L"ocamlBase.txt", camlDirPath est le chemin absolu de la distribution Caml Light ou OCaml
{
    wstring s;
    if (getConfigFolderPath(s))
    {
        s += camlBase;
        //infoMessage(wstring(L"writeCamlPath: " + s));       
        fstream fs(toBytes(s).c_str(), fstream::out);
        if (!fs.fail())
        {
            fs << fromUnicode(camldirPath, ENCODING_UTF8) << endl;
            fs.close();
        }
    }
}
wstring setCamlPath(void)
{
    wstring message = L"Le dialogue suivant devrait permettre de naviguer vers un dossier Caml valide ou d'annuler.";
    QString prompt = "Emplacement de Caml ?";
    ::infoMessage(message);
    QFileDialog fdlg(mainFrame, prompt, QDir::homePath());
    QStringList fileNames;
    fdlg.setFileMode(QFileDialog::Directory);
    fdlg.setFilter(QDir::AllEntries | QDir::Hidden);
    if (fdlg.exec()){
        fileNames = fdlg.selectedFiles();
        fdlg.close();
    }
    return (fileNames.isEmpty() ? "" : fileNames[0]).toStdWString();
}
bool resetCamlPath(const QString& camlbase)
{
    wstring s;
    if (getConfigFolderPath(s))
    {
        void deleteFile(const wstring& fileName);
        s += camlbase.toStdWString();
        QFileInfo fi = QString::fromStdWString(s);
        if (fi.exists())
        {
            QString distrib = camlbase == "camlbase.txt" ? "Caml Light" : "OCaml";
            QString message = "Abandonner la distribution " + distrib + " actuelle ?";
            if (::yesnoMessage(message.toStdWString()))
            {
                deleteFile(s);
            }
            return true;
        }
    }
    return false;
}

bool resetCamlPaths()
{
    bool b1 = resetCamlPath("camlbase.txt");
    bool b2 = resetCamlPath("ocamlbase.txt");
    return b1 || b2;
}
/*
#ifdef Q_OS_LINUX
char byte64(const char* path)
{
    unsigned char c[5] = {0};
    FILE* f = fopen(path, "rb");
    if (!f) return 0;
    if (fread(c, 5, 1, f) == 1)
    {
        fclose(f);
        return c[4];
    }
    return 0;
}
#endif

bool isExecutable(const QString& runtimePath)
{
    FILE* f = fopen(runtimePath.toStdString().c_str(), "rb");
    if (!f) return false;
    char c[5];
    if (fread(c, 5, 1, f) == 1)
    {
        fclose(f);
#ifdef Q_OS_LINUX
        if (c[0] != 0x7f || c[1] != 0x45 || c[2] != 0x4c || c[3] != 0x46 || c[4] != byte64("/bin/sh"))
#else
            uint32_t m = *(uint32_t*)c;
        if (m != 0xfeedface && m != 0xfeedfacf && m != 0xcefaedfe && m != 0xcffaedfe)
#endif
        {
            errorMessage(L"Format de fichier binaire incorrect");
            return true;
        }
    }
    return true;
}

#ifdef Q_OS_LINUX
bool isElf64leExecutable(const char* path)
{
    unsigned char c[6] = {0};
    FILE* f = fopen(path, "rb");
    if (!f) return false;
    if (fread(c, 6, 1, f) == 1)
    {
        fclose(f);
        return c[0] == 0x7f && c[1] == 0x45 && c[2] == 0x4c && c[3] == 0x46 && c[4] == 0x02 && c[5] == 0x01;
    }
    return false;
}
#else
bool isMac64Executable(const char* path)
{
    unsigned char c[4] = {0};
    FILE* f = fopen(path, "rb");
    if (!f) return false;
    if (fread(c, 4, 1, f) == 1)
    {
        fclose(f);
        return c[0] == 0xcf && c[1] == 0xfa && c[2] == 0xed && c[3] == 0xfe;
    }
    return false;
}
#endif
*/
CProcess* startOCaml(void)
{ 
    wstring ocamlpath = getCamlPath(L"ocamlbase.txt"); 
    QString camlPath = QString::fromStdWString(ocamlpath);
/*
    QFileInfo fi1 = camlPath + "/bin/ocamlrun";
    QFileInfo fi2 = camlPath + "/bin/ocaml";
    if (!fi1.exists() || !fi2.exists() || !fi1.isExecutable() || !isExecutable(camlPath + "/bin/ocamlrun"))
    {
        return NULL;
    }
*/

    struct subpaths dir_subpaths(const wstring &p, bool camlnew=false);
    struct subpaths sp = dir_subpaths(ocamlpath);
    if (!sp.ocaml || sp.ocamllib == L"" || sp.stublibs == L"" || sp.findlib == L"") sp = dir_subpaths(ocamlpath, true);
    QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
    env.insert("OCAMLLIB", QString::fromStdWString(sp.ocamllib));
    //infoMessage(wstring(L"OCAMLLIB: ") + env.value("OCAMLLIB").toStdWString());
    QString camlldlibrarypath = QString::fromStdWString(sp.stublibs);
    env.insert("CAML_LD_LIBRARY_PATH", camlldlibrarypath);
    //infoMessage(wstring(L"CAML_LD_LIBRARY_PATH: ") + env.value("CAML_LD_LIBRARY_PATH").toStdWString());   
    if (sp.findlib != L""){
        wstring cfp;
        if (getConfigFolderPath(cfp)){
            cfp += L"findlib.conf";
            env.insert("OCAMLFIND_CONF", QString::fromStdWString(cfp));
            //infoMessage(wstring(L"OCAMLFIND_CONF: ") + env.value("OCAMLFIND_CONF").toStdWString());
        }
    }
    QString p = camlPath + "/bin:" + env.value("PATH"); 
    env.insert("PATH", p);
    //env.insert("PATH", camlPath + "/bin:/usr/local/bin:/usr/bin:/bin");
    //infoMessage(wstring(L"PATH: ") + env.value("PATH").toStdWString());        
    QStringList args = (QStringList(camlPath + (sp.ocwin ?"/bin/ocaml.exe" : "/bin/ocaml")) << "-I" << QString::fromStdWString(configFolderPath()))<< "-I" << applicationDirPath() + "/ocamlgraph";
    QDir::setCurrent(applicationDirPath());
    CProcess* process = new CProcess();
    process->setProcessEnvironment(env);
    QString cmdl = camlPath + (sp.ocwin ?"/bin/ocamlrun.exe " : "/bin/ocamlrun ") + args.join(" ");
    //infoMessage(L"démarrage d'OCaml...");
    //infoMessage(wstring(L"LIGNE DE COMMANDE: ") + cmdl.toStdWString());
    process->start(camlPath + (sp.ocwin ?"/bin/ocamlrun.exe" : "/bin/ocamlrun"), args);
    return process;            
}

CProcess* startCamllight(void)
{
    QString camlPath = QString::fromStdWString(getCamlPath(L"camlbase.txt"));
    QString camlLib = camlPath + "/lib/caml-light";
    QString toplevel = camlLib + "/caml_all";
    QFileInfo fi = toplevel;
    if (!fi.exists()) {
    	toplevel = camlLib + "/caml_all.exe";
    	camlLib = camlLib.mid(5,1) + ":" + camlLib.mid(6);
    }
//   QFileInfo fi = toplevel;
//   if (!fi.exists() || !fi.isExecutable() || !isExecutable(toplevel))
//   {
//       return NULL;
//   }
     QProcessEnvironment env = QProcessEnvironment::systemEnvironment();        
     //env.insert("PATH", camlPath + "/bin:/usr/local/bin:/usr/bin:/bin:");
     env.insert("PATH", camlPath + "/bin:" + env.value("PATH"));
     //infoMessage(wstring(L"PATH: ") + env.value("PATH").toStdWString());
     env.insert("CAMLLIGHTDIR", camlPath);
            
     QString locale = QLocale::system().name().mid(0, 2);
     QStringList args = (QStringList() << "-stdlib" << camlLib << "-lang" << locale);
     QDir::setCurrent(applicationDirPath());

     CProcess* process = new CProcess();
     process->setProcessEnvironment(env);
     QString cmdl = toplevel + " " + args.join(" ");
     //infoMessage(L"démarrage de Caml Light...");
     //infoMessage(wstring(L"LIGNE DE COMMANDE CAML LIGHT: ") + cmdl.toStdWString());
     process->start(toplevel, args);
     return process;
}

CProcess* startCaml(bool ocaml)
{
    wstring path = ocaml ? getCamlPath(L"ocamlbase.txt") : getCamlPath(L"camlbase.txt");
    wstring caml = ocaml ? L"OCaml" : L"Caml light";
    if (path != L"") {
        CProcess* camlProcess = ocaml ? startOCaml() : startCamllight();
        if (camlProcess){
            return camlProcess;
        }
        else {
            infoMessage(caml + L": démarrage impossible dans " + path);
        }
    }
    else {
        infoMessage(L"La distribution " + caml +
                    L" n'a pas \u00E9t\u00E9 choisie.\nRecommencer apr\u00E8s avoir \u00E0 partir de l'explorateur de fichiers"
                    " fait  glisser son dossier dans la fen\u00EAtre de WinCaml ou apr\u00E8s avoir choisi ce dossier \u00E0 partir"
                    " du menu 'Caml->Nouvelle distribution...'");
    }
    return NULL;
 }
 

void camllightHelp()
{
#ifdef Q_OS_MAC
    QStringList args = (QStringList() << "-a" << "safari.app" << applicationDirPath() + "/doc/man-caml/index.html");
    QProcess::startDetached("open", args);
#else
#ifdef Q_OS_LINUX
    QStringList args = (QStringList() << qApp->applicationDirPath() + "/doc/man-caml/index.html");
    QProcess::startDetached("/usr/bin/firefox", args);
#endif
#endif
}
void ocamlHelp()
{
#ifdef Q_OS_MAC
    QStringList args = (QStringList() << "-a" << "safari.app" << applicationDirPath() + "/doc/htmlman/index.html");
    QProcess::startDetached("open", args);
#else
#ifdef Q_OS_LINUX
    QStringList args = (QStringList() << qApp->applicationDirPath() + "/doc/htmlman/index.html");
    QProcess::startDetached("/usr/bin/firefox", args);
#endif
#endif
}
#endif

#ifdef Q_OS_WIN
#include "../Win/Paths.cpp"
#endif
