/*
 Program WinCaml: Graphical User Interface
 for interactive use of Caml-Light and Ocaml.
 Copyright (C) 2005-2024 Jean Mouric 35700 Rennes France
 email: jean.mouric@orange.fr

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

// File Win/Paths.cpp  (aussi incluse par Qt/Path.cpp ou wx/Paths.cpp)

#ifdef _WINDOWS_PLATFORM_
#include "Win.h"
#else
#include <string> // dj inclus dans Win.h
#include <algorithm>  // dj inclus dans Win.h
#endif
#include <fstream>
#include <sstream>
#include <shlobj.h>
#include "Shlwapi.h"
#include <filesystem> // ncessite -std=c++17

#if defined (Q_OS_WIN) || defined (__WXMSW__)
bool getConfigFolderPath(wstring& path)
{
#ifdef Q_OS_WIN
    wchar_t szPath[MAX_PATH];
#else
    WCHAR szPath[MAX_PATH];
#endif
    if (::SHGetFolderPath(NULL, CSIDL_LOCAL_APPDATA, NULL, 0, szPath) == S_OK)
    {
        path = szPath;
        path += L"\\" + appName + L"\\";
        return true;
    }
    return false;
}
wstring applicationDirectoryPath()
{
#ifdef Q_OS_WIN
    wchar_t appPath[MAX_PATH];
#else
    WCHAR appPath[MAX_PATH];
#endif
    ::GetModuleFileName(0, appPath, sizeof(appPath) - 1);
    wstring appDirPath = appPath;
    return appDirPath.substr(0, appDirPath.find_last_of(L"\\"));
}

#ifdef QT4_7
typedef BOOL(WINAPI *LPFN_ISWOW64PROCESS) (HANDLE, BOOL*);
LPFN_ISWOW64PROCESS fnIsWow64Process;

static bool IsWow64()
{
    BOOL bIsWow64 = FALSE;
    //IsWow64Process is not available on all supported versions of Windows.
    //Use GetModuleHandle to get a handle to the DLL that contains the function
    //and GetProcAddress to get a pointer to the function if available.
    fnIsWow64Process = (LPFN_ISWOW64PROCESS) ::GetProcAddress(::GetModuleHandle(TEXT("kernel32")), "IsWow64Process");
    if (NULL != fnIsWow64Process)
    {
        if (!fnIsWow64Process(::GetCurrentProcess(), &bIsWow64))
        {
            //handle error
        }
    }
    return bIsWow64 == TRUE;
}

static bool architecture64()
{
    return IsWow64();
}
#else
static bool architecture64()
{
    SYSTEM_INFO sysInfo;
    ::GetNativeSystemInfo(&sysInfo);
    return sysInfo.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_AMD64 /*|| sysInfo.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_ARM64*/;
}
#endif
#endif

extern CMDIFrame* mainFrame;

bool getConfigFolderPath(wstring& path); // dfinie dans Win.cpp, sinon ci-dessus
//struct subpaths dfinie dans Win.h ou Qt.h
//struct subpaths dir_subpaths(const std::filesystem::path& directory, bool camlnew = false); // dfinie dans dire.cpp
struct subpaths dir_subpaths(const wstring &p, bool camlnew=false);  // dir_subpaths dfinie dans dire.cpp
void infoMessage(const wstring & msg); // dfinie dans Win.cpp ou Qt.cpp
wstring fromBytes(const string& str); // utf-8 vers utf-16 , dfinie dans Win.cpp ou Qt.cpp
string toBytes(const wstring& wstr); // utf-16 vers utf-8 , dfinie dans Win.cpp ou Qt.cpp
void deleteFile(const wstring& fileName); // dfinie dans Win.cpp ou Qt.cpp
std::wstring wslpath(std::wstring path); // dfinie dans dire.cpp
wstring applicationDirectoryPath(void); //dfinie dans Win.cpp ou ci-dessus

wstring configFolderPath(void)
{
    wstring path = L"";
    if (getConfigFolderPath(path)) replace(path.begin(), path.end(), '\\', '/');
    return path;
}

wstring getCamlPath(bool ocaml)
{
    wstring cfp;
    string camlPath = "";
    if (getConfigFolderPath(cfp))
    {
        cfp += ocaml ? L"ocamlbase.txt" : L"camlbase.txt";
        fstream fs(cfp.c_str(), fstream::in);
        if (!fs.fail())
        {
            char c;
            while (fs.get(c) && c != L'\r' && c != L'\n')
                camlPath += c;

            fs.close();
        }
    }
    return fromBytes(camlPath);
}

wstring setCamlPath(void)
{
    //dismissed = false;
    wstring message = L"Le dialogue suivant devrait permettre de naviguer vers un dossier Caml valide ou d'annuler.";
    ::infoMessage(message);
    WCHAR path[MAX_PATH];
    memset(path, 0, sizeof(path));
    BROWSEINFO bi = {0};
    bi.lpszTitle =  L"Emplacement de Caml ?";
#ifdef _WINDOWS_PLATFORM_
    bi.hwndOwner = mainFrame->hWnd;
#endif
#ifdef __WXMSW__
    bi.hwndOwner = (HWND)mainFrame->GetHWND();
#endif
#ifdef Q_OS_WIN
    bi.hwndOwner = (HWND)mainFrame->winId();
#endif
    //bi.ulFlags |= BIF_NEWDIALOGSTYLE;
    bi.ulFlags |= BIF_USENEWUI | BIF_BROWSEINCLUDEFILES | BIF_BROWSEINCLUDEURLS;
    LPITEMIDLIST pidl = SHBrowseForFolder(&bi);
    if (pidl != 0)
    {
        SHGetPathFromIDList (pidl, path);
        IMalloc* p = NULL;
        if ( SUCCEEDED(SHGetMalloc(&p)))
        {
            p->Free (pidl);
            p->Release();
        }
        return path;
    }
    //dismissed = true;
    return L"";
}

void writeCamlPath(wstring camlPath, wstring camlBase)
{
    wstring s;
    if (getConfigFolderPath(s))
    {
        s += camlBase;
        fstream fs(toBytes(s).c_str(), fstream::out);
        if (!fs.fail())
        {
            fs << toBytes(camlPath) << endl;
            fs.close();
        }
    }
}

bool resetCamlPath(const wstring& camlbase)
{
    wstring s;
    if (getConfigFolderPath(s))
    {
        s += camlbase;
        if (PathFileExists(s.c_str()))
        {
            wstring distrib = camlbase == L"camlbase.txt" ? L"Caml Light" : L"OCaml";
            wstring message = L"Abandonner la distribution " + distrib + L" actuelle ?";
            if (::yesnoMessage(message))
            {
                deleteFile(s);
            }
            return true;
        }
    }
    return false;
}


bool resetCamlPaths()
{
    bool b1 = resetCamlPath(L"camlbase.txt");
    bool b2 = resetCamlPath(L"ocamlbase.txt");
    return b1 || b2;
}

bool architecture64();

bool isSixtyFour(const wstring& path)
{
    unsigned char t[2];
    FILE* f = _wfopen((wchar_t*)path.c_str(), L"rb");
    if (f)
    {
        int i = 0;
        while (fread(&t, 2, 1, f) == 1 && (t[0] != 0x50 || t[1] != 0x45))
        {
            i++;
        }
        if (fread(&t, 2, 1, f) == 1 && fread(&t, 2, 1, f) == 1)
        {
            bool b = t[0] + t[1] * 0x100 == 0x8664;
            return b;
        }
    }
    return architecture64();
}

int isCygwin(const wstring& path)
{
    char* q = (char*)"cygwin1.dll";
    char* p = q;
    char c;
    FILE* f = _wfopen(path.c_str(), L"rb");
    if (!f) return 0;
    while (fread(&c, 1, 1, f) == 1 && *p != 0)
    {
        if (*p != c) p = q; else p++;
    }
    fclose(f);
    return *p == 0;

}

wstring ocamlCommandLine(wstring ocamlPath, wstring& dirPath, string& cygwin)
{
    if (ocamlPath == L"")
    {
        return L"";
    }

    struct subpaths sp = dir_subpaths(ocamlPath);
    if (!sp.ocaml || sp.ocamllib == L"" || sp.stublibs == L"" || sp.findlib == L"")
        sp = dir_subpaths(ocamlPath, true);
    if (sp.ocaml && !sp.ocwin)
    {
        wstring cmdl = LR"(wsl -- ocamldir=")" + wslpath(ocamlPath) + LR"(";
export PATH=$ocamldir/bin:$PATH;
export DISPLAY=:0.0;
export OCAMLLIB=")" + sp.ocamllib + LR"("
export CAML_LD_LIBRARY_PATH=")" + sp.stublibs + LR"("
export OCAMLFIND_CONF=")" + LR"(/mnt/c)" + configFolderPath().substr(2) + LR"(findlib.conf"
"$ocamldir"/bin/ocamlrun "$ocamldir"/bin/ocaml -I )" + wslpath(configFolderPath()) + LR"( -I )" + LR"("$HOME"/WinCaml7/ocamlgraph)";
       // infoMessage(wstring(L"LIGNE DE COMMANDE: ") + cmdl);
        return cmdl;
    }

    wstring cmdl;
    wstring camlrun = ocamlPath + L"\\bin\\ocamlrun.exe";
    bool sixtyfour = isSixtyFour(camlrun);
    //wstring libgraphDirPath = dirPath + L"\\ocaml\\windows";
    //sixtyfour ? libgraphDirPath += L"\\ocaml64" : libgraphDirPath += L"\\ocaml32";
    wstring libgraphDirPath = dirPath + L"\\ocamlgraph";

    wstring ocamlLibPath = ocamlPath + L"\\lib\\ocaml";

    if (isCygwin(camlrun))
    {
        string s1 = toBytes(ocamlPath);
        std::replace(s1.begin(), s1.end(), '\\', '/');
        char c = 'c';
        if (s1.length() > 2 && s1[1] == char(':'))
        {
            c = s1[0];
            s1 = string("/cygdrive/").append(s1.erase(1, 1));
        }
        string s2 = toBytes(libgraphDirPath);
        std::replace(s2.begin(), s2.end(), '\\', '/');
        if (s2.length() > 2 && s2[1] == char(':'))
        {
            s2 = string("/cygdrive/").append(s2.erase(1, 1));
        }
        cygwin = string(1, c) + ":\\cygwin";
        if (sixtyfour) cygwin += "64";
        wstring cygwinW = fromBytes(cygwin);
        cmdl = cygwinW + L"\\bin\\bash.exe -c 'ocamlrun ocaml -I \"$LIBGRAPHDIRPATH\" -I \"$OCAMLPATH/lib/ocaml/stublibs\"'";
        ::SetEnvironmentVariable(wstring(L"OCAMLLIB").c_str(), fromBytes(s1 + "/lib/ocaml").c_str());
        ::SetEnvironmentVariable(wstring(L"PATH").c_str(), (ocamlPath + L"\\bin;" + cygwinW + L"\\bin;C:\\Windows\\System32;C:\\Windows").c_str());
        ::SetEnvironmentVariable(wstring(L"CAML_LD_LIBRARY_PATH").c_str(), fromBytes(s1 + "/lib/ocaml/stublibs").c_str());
        ::SetEnvironmentVariable(wstring(L"OCAMLPATH").c_str(), fromBytes(s1).c_str());
        ::SetEnvironmentVariable(wstring(L"LIBGRAPHDIRPATH").c_str(), fromBytes(s2).c_str());
    }
    else
    {
        wstring ocamlrun = L"\"" + ocamlPath + L"\\bin\\ocamlrun.exe\"" + L" \"" + ocamlPath + L"\\bin\\ocaml.exe\"";
        std::wstring cfp = configFolderPath();
        cmdl = ocamlrun + L" -I \"" + cfp + L"\" -I \"" + libgraphDirPath + L"\"";
        ::SetEnvironmentVariable(wstring(L"PATH").c_str(), (ocamlPath + L"\\bin;C:\\Windows\\System32;C:\\Windows").c_str());
        ::SetEnvironmentVariable(wstring(L"OCAMLFIND_CONF").c_str(), (cfp + L"/findlib.conf").c_str());
        ::SetEnvironmentVariable(wstring(L"OCAMLLIB").c_str(), sp.ocamllib.c_str());
        ::SetEnvironmentVariable(wstring(L"CAML_LD_LIBRARY_PATH").c_str(), sp.stublibs.c_str());
    }
    //infoMessage(wstring(L"LIGNE DE COMMANDE: ") + cmdl);
    return cmdl;
}

wstring camllightCommandLine(const wstring& camlPath, string& cygwin)
{
    if (camlPath == L"")
    {
        return L"";
    }
    wstring camldirPath = camlPath;
    struct subpaths sp = dir_subpaths(camldirPath);
    if (sp.camllight && !sp.clwin)
    {
        camldirPath = wslpath(camldirPath);
        wstring toplevel = camldirPath + L"\\lib\\caml-light\\caml_all";
        wstring cmdl = L"wsl -- export DISPLAY=:0.0;\"" + camldirPath + L"/lib/caml-light/caml_all\"" + L" -stdlib \"" + camldirPath + L"/lib/caml-light\"" + L" -lang \"fr\"";
        //infoMessage(wstring(L"LIGNE DE COMMANDE: ") + cmdl);
        return cmdl;
    }
    wstring toplevel = camlPath + L"\\lib\\caml-light\\caml_all.exe";
    if (!PathFileExists(toplevel.c_str()))
    {
        return L"";
    }
    wstring cmdl;
    if (isCygwin(toplevel))
    {
        string camlPath1 = toBytes(camlPath);
        std::replace(camlPath1.begin(), camlPath1.end(), '\\', '/');
        char c = 'c';
        if (camlPath1.length() > 2 && camlPath1[1] == char(':'))
        {
            c = camlPath1[0];
            camlPath1 = string("/cygdrive/").append(camlPath1.erase(1, 1));
        }

        wstring caml_allPath = camlPath + L"\\lib\\caml-light\\caml_all.exe";
        string archi = isSixtyFour(caml_allPath) ? "64" : "";
        cygwin = string(1, c) + ":\\cygwin" + archi;
        wstring cygwinW = fromBytes(cygwin);

        cmdl = cygwinW + L"\\bin\\bash -c '\"$CAMLLIGHTDIR/lib/caml-light/caml_all\" -stdlib \"$CAMLLIGHTDIR/lib/caml-light\" -lang \"fr\"'";
        ::SetEnvironmentVariable(wstring(L"PATH").c_str(), (camlPath + L"\\caml-light\\bin;" + cygwinW + L"\\bin;C:\\Windows\\System32;C:\\Windows").c_str());
        ::SetEnvironmentVariable(wstring(L"CAMLLIGHTDIR").c_str(), fromBytes(camlPath1).c_str());
    }
    else
    {
        ::SetEnvironmentVariable(wstring(L"PATH").c_str(), (camlPath + L"\\bin;C:\\Windows\\System32;C:\\Windows").c_str());
        ::SetEnvironmentVariable(wstring(L"CAMLLIGHTDIR").c_str(), camlPath.c_str());
        cmdl = L"\"" + toplevel +L"\" -stdlib \"" + camlPath + L"\\lib\\caml-light\"" + L" -lang \"fr\"";
    }
    //infoMessage(wstring(L"LIGNE DE COMMANDE: ") + cmdl);
    return cmdl;
}

CProcess* startCaml(bool ocaml)
{
    wstring path = getCamlPath(ocaml);
    wstring caml = ocaml ? L"OCaml" : L"Caml Light";
    if (path != L"")
    {
        if (!std::filesystem::exists(path) || !std::filesystem::is_directory(path)){
            infoMessage(path + L" n'est pas en ligne ou n'est pas un dossier !");
            return NULL;
        }
        wstring startcaml = wstring(L":\n D\u00E9marrage dans ");
        string cygwin = "";
        wstring dirPath = applicationDirectoryPath();
        wstring cmdl = ocaml ? ocamlCommandLine(path, dirPath, cygwin) : camllightCommandLine(path, cygwin);
        if (cmdl != L"")
        {
            CProcess* camlProcess = new CProcess();
            if (camlProcess->start(cmdl))
            {
                if (cygwin != "") camlProcess->setPid(cygwin);
                return camlProcess;
            }
            else {
                infoMessage(caml + L": d\u00E9marrage impossible dans " + path);
            }
        }
    }
    else {
        infoMessage(L"La distribution " + caml +
                    L" n'a pas \u00E9t\u00E9 choisie.\nRecommencer apr\u00E8s avoir \u00E0 partir de l'explorateur de fichiers"
                    " fait  glisser son dossier dans la fen\u00EAtre de WinCaml ou apr\u00E8s avoir choisi ce dossier \u00E0 partir"
                    " du menu 'Caml->Nouvelle distribution...'");
    }
    return NULL;
}

void camllightHelp()
{
    wstring appDirPath = applicationDirectoryPath();
    wstring docPath = appDirPath + L"\\doc\\man-caml\\index.html";
    ::ShellExecute(NULL, L"open", docPath.c_str(), NULL, appDirPath.c_str(), SW_SHOW);
}

void ocamlHelp()
{
    wstring appDirPath = applicationDirectoryPath();
    wstring docPath = appDirPath;
    docPath += L"\\doc\\htmlman\\index.html";
    if (32 >= (size_t)::ShellExecute(NULL, L"open", docPath.c_str(), NULL, appDirPath.c_str(), SW_SHOW))
    {
        if (32 >= (size_t)::ShellExecute(NULL, L"open", L"C:\\Program Files\\Objective Caml\\htmlman\\index.html", NULL, L"C:\\Program Files\\Objective Caml\\htmlman", SW_SHOW))
            ::ShellExecute(NULL, L"open", L"C:\\Program Files (x86)\\Objective Caml\\htmlman\\index.html", NULL, L"C:\\Program Files (x86)\\Objective Caml\\htmlman", SW_SHOW);
    }
}
